@@ -381,6 +381,29 @@ TEST(xpath_parse_oom_propagation)
 	}
 }
 
+static std::basic_string<char_t> rep(const std::basic_string<char_t>& base, size_t count)
+{
+	std::basic_string<char_t> result;
+	result.reserve(base.size() * count);
+
+	for (size_t i = 0; i < count; ++i)
+		result += base;
+
+	return result;
+}
+
+TEST(xpath_parse_depth_limit)
+{
+	const size_t limit = 1500;
+
+	CHECK_XPATH_FAIL((rep(STR("("), limit) + STR("1") + rep(STR(")"), limit)).c_str());
+	CHECK_XPATH_FAIL((STR("(id('a'))") + rep(STR("[1]"), limit)).c_str());
+	CHECK_XPATH_FAIL((STR("/foo") + rep(STR("[1]"), limit)).c_str());
+	CHECK_XPATH_FAIL((STR("/foo") + rep(STR("/x"), limit)).c_str());
+	CHECK_XPATH_FAIL((STR("1") + rep(STR("+1"), limit)).c_str());
+	CHECK_XPATH_FAIL((STR("concat(") + rep(STR("1,"), limit) + STR("1)")).c_str());
+}
+
 TEST_XML(xpath_parse_location_path, "<node><child/></node>")
 {
 	CHECK_XPATH_NODESET(doc, STR("/node")) % 2;
